package soot.javaToJimple.ppa;

import polyglot.types.ArrayType;
import polyglot.types.Type;
import soot.javaToJimple.ppa.jj.types.MagicClass;

public class PPATypeMerger {

	@SuppressWarnings("null")
	private TypeFact getSubType(TypeFact oldTuple, TypeFact newTuple) {
		TypeFact tuple = null;
		Type oldNewType = oldTuple.getNewType();
		Type newNewType = newTuple.getNewType();

		if (TypeFactUtil.isMagicClassKind(oldNewType) && TypeFactUtil.isMagicClassKind(newNewType)) {
			tuple = null;
		} else if (oldNewType.isSubtype(newNewType)) {
			tuple = oldTuple;
		} else if (newNewType.isSubtype(oldNewType)) {
			tuple = newTuple;
		}

		return tuple;
	}

	private TypeFact getSuperType(TypeFact oldTuple, TypeFact newTuple) {
		TypeFact tuple = null;
		Type oldNewType = oldTuple.getNewType();
		Type newNewType = newTuple.getNewType();

		if (oldNewType.isSubtype(newNewType)) {
			tuple = newTuple;
		} else if (newNewType.isSubtype(oldNewType)) {
			tuple = oldTuple;
		}

		return tuple;
	}

	private boolean isPrimitiveInvolved(Type oldType, Type newType) {
		return oldType.isPrimitive() || newType.isPrimitive();
	}

	private boolean isArrayInvolved(Type oldType, Type newType) {
		return oldType.isArray() || newType.isArray();
	}

	/**
	 * <p>
	 * Merge two type facts.
	 * </p>
	 * 
	 * <p>
	 * Some characteristics of the merge operation:
	 * <ul>
	 * <li>Parameters are not interchangeable: merge(tuple1,tuple2) does not necessarily produce
	 * the same result as merge(tuple2,tuple1).</li>
	 * <li>The operation is stable: as long as the old tuple is always favored over the new one
	 * when there is a conflict: this means that merging conflicting type facts will eventually
	 * converge (no infinite loop).</li>
	 * </ul>
	 * </p>
	 * 
	 * @param oldTuple
	 * @param newTuple
	 * @return
	 */
	public TypeFact merge(TypeFact oldTuple, TypeFact newTuple) {
		TypeFact mergedTuple = null;
		Type oldNewType = oldTuple.getNewType();
		Type newNewType = newTuple.getNewType();
		TypeFact subTuple = getSubType(oldTuple, newTuple);
		TypeFact superTuple = getSuperType(oldTuple, newTuple);
		// oldNewType and newNewType should never be unknown magic class since this is not an
		// improvement!

		// when should we erase type?
		// when should we only add type information?
		// ex: unknown field is ok!
		// local dec: never!
		//

		if (oldNewType.equals(newNewType)) {
			mergedTuple = oldTuple;
		} else if (isPrimitiveInvolved(oldNewType, newNewType)) {
			mergedTuple = mergePrimitives(oldTuple, newTuple);
		} else if (isArrayInvolved(oldNewType, newNewType)) {
			mergedTuple = mergeArray(oldTuple, newTuple);
		} else if (subTuple != null) {
			mergedTuple = mergeHierarchyRelatedTuples(subTuple, superTuple);
		} else {
			mergedTuple = mergeUnrelatedTuples(oldTuple, newTuple);
		}

		return mergedTuple;
	}

	private TypeFact mergeArray(TypeFact oldTuple, TypeFact newTuple) {
		TypeFact mergeTuple = null;

		if (oldTuple.getNewType().isArray() && !newTuple.getNewType().isArray()) {
			mergeTuple = oldTuple;
		} else if (!oldTuple.getNewType().isArray() && newTuple.getNewType().isArray()) {
			mergeTuple = newTuple;
		} else {
			// Both are arrays, check for anonymous or unsafe!
			Type oldType = ((ArrayType) oldTuple.getNewType()).base();
			Type newType = ((ArrayType) newTuple.getNewType()).base();
			if (TypeFactUtil.isTypeSafer(oldType, newType)) {
				mergeTuple = oldTuple;
			} else if (TypeFactUtil.isTypeSafer(newType, oldType)) {
				mergeTuple = newTuple;
			} else {
				mergeTuple = oldTuple;
			}
		}

		return mergeTuple;
	}

	private TypeFact mergeHierarchyRelatedTuples(TypeFact subTuple, TypeFact superTuple) {
		TypeFact mergedTuple = null;

		// ? = Object
		// String = ?
		// Should not compile!
		if (subTuple.getKind() == TypeFact.CHILD_OR_EQUALS
				&& superTuple.getKind() == TypeFact.SUPER_OR_EQUALS) {
			System.err.println("Potential conflict detected!");
			mergedTuple = subTuple;
		} else if (superTuple.getKind() == TypeFact.SUPER_OR_EQUALS) {
			mergedTuple = superTuple;
		} else {
			mergedTuple = subTuple;
		}

		return mergedTuple;
	}

	private TypeFact mergePrimitives(TypeFact oldTuple, TypeFact newTuple) {
		Type oldNewType = oldTuple.getNewType();
		Type newNewType = newTuple.getNewType();
		TypeFact mergedTuple = null;

		if (oldNewType.isPrimitive() && !newNewType.isPrimitive()) {
			// System.err.println("Potential conflict detected: old is primitive, new is not.");

			// If old = primitive, old is primitive.
			// If primitive = old, old is not primitive!

			mergedTuple = oldTuple;
		} else if (newNewType.isPrimitive() && !oldNewType.isPrimitive()) {

			// Same here!
			// System.err.println("Potential conflict detected: new is primitive, old is not");
			mergedTuple = newTuple;
		} else if (newNewType.isPrimitive() && oldNewType.isPrimitive()) {
			// TODO Do something similar to subTuple but for primitives!
			mergedTuple = oldTuple;
		}

		return mergedTuple;
	}

	private TypeFact mergeUnrelatedTuples(TypeFact oldTuple, TypeFact newTuple) {
		TypeFact mergedTuple = null;

		// System.err.println("Potential conflict detected: not same hiearchy");

		if (oldTuple.getNewType() instanceof MagicClass) {
			MagicClass mc = (MagicClass) oldTuple.getNewType();
			mc.addRelatedTypes(newTuple.getNewType());
		}

		if (newTuple.getNewType() instanceof MagicClass) {
			MagicClass mc = (MagicClass) newTuple.getNewType();
			mc.addRelatedTypes(oldTuple.getNewType());
		}

		mergedTuple = oldTuple;

		return mergedTuple;
	}

}
