package soot.javaToJimple.ppa;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import polyglot.ast.Node;
import polyglot.types.ArrayType;
import polyglot.types.MethodInstance;
import polyglot.types.NullType;
import polyglot.types.ReferenceType;
import polyglot.types.Type;
import polyglot.util.Position;
import soot.javaToJimple.ppa.jj.types.MagicClass;
import soot.javaToJimple.ppa.jj.types.MagicClassArray;
import soot.javaToJimple.ppa.jj.types.MagicPrimitive;

public class TypeFactUtil {

	public final static int ANON_MAGIC = 0;

	public final static int ANON_MAGIC_ARRAY = 1;

	public final static int MAGIC_MAGIC = 2;

	public final static int MAGIC_SUPER = 3;

	public final static int MAGIC_INTER = 4;

	public final static int NO_MAGIC = 5;

	/**
	 * 
	 * @param objects
	 * @return True if one of the object is an anoymous unknown class.
	 */
	public static boolean containsAnonymous(List<? extends Object> objects) {
		boolean containsUnknown = false;

		for (Object obj : objects) {
			if (isAnonymousMagicClass(obj)) {
				containsUnknown = true;
				break;
			}
		}

		return containsUnknown;
	}

	public static MagicClass getFirstMagicClassInHierarchy(ReferenceType container) {
		Type type = container;
		while (type != null && !(type instanceof MagicClass)) {
			ReferenceType ref = type.toReference();
			if (ref != null) {
				type = ref.superType();
			} else {
				type = null;
			}
		}

		if (!(type instanceof MagicClass)) {
			type = null;
		}

		return (MagicClass) type;
	}

	public static List<PPAIndex> getIndexes(Object node) {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		if (node != null && node instanceof PPANode) {
			PPANode sn = (PPANode) node;
			indexes = sn.getIndexes();
		}
		return indexes;
	}

	/**
	 * @param node
	 * @return The main index of a PPANode, null if node is not an instance of PPANode
	 */
	public static PPAIndex getMainIndex(Object node) {
		PPAIndex index = null;
		if (node != null && node instanceof PPANode) {
			PPANode sn = (PPANode) node;
			index = sn.getMainIndex();
		}
		return index;
	}

	/**
	 * 
	 * @param position
	 * @return A hashcode, relatively robust, for ast node position.
	 */
	public static int hashCode(Position position) {
		return position.nameAndLineString().hashCode() + position.column() + position.endColumn();
	}

	public static boolean hasMagic(MethodInstance mi) {
		boolean hasMagic = isMagicClassKind(mi.container());
		hasMagic = hasMagic || isMagicClassKind(mi.returnType());
		if (!hasMagic) {
			for (Object param : mi.formalTypes()) {
				if (isMagicClassKind(param)) {
					hasMagic = true;
					break;
				}
			}
		}

		return hasMagic;
	}

	/**
	 * 
	 * @param type
	 * @return True if the object has a MagicClass in its hierarchy (either through super types or
	 *         implemented interfaces).
	 */
	public static boolean hasMagicInHierarchy(ReferenceType type) {
		if (type == null) {
			return false;
		}

		boolean hasMagic = type instanceof MagicClass;

		if (!hasMagic) {
			Type superType = type.superType();

			// First check super types
			if (superType != null && superType.toReference() != null) {
				hasMagic = hasMagicInHierarchy(superType.toReference());
			}

			// Then check interfaces
			if (!hasMagic) {
				List types = type.interfaces();
				for (Object typeObject : types) {
					Type tempType = (Type) typeObject;
					ReferenceType rType = tempType.toReference();
					if (rType != null) {
						hasMagic = hasMagicInHierarchy(rType);
						if (hasMagic) {
							break;
						}
					}
				}
			}
		}

		return hasMagic;
	}

	/**
	 * 
	 * @param tuple
	 * @return True if oldType != newType and newType != NullType
	 */
	public static boolean hasNewTypeInfo(TypeFact tuple) {
		Type newType = tuple.getNewType();
		return !newType.equalsImpl(tuple.getOldType()) && !(newType instanceof NullType);
	}

	/**
	 * 
	 * @param left
	 * @param right
	 * @param l
	 * @param r
	 * @return True if right is unsafe and left is safe or if right is unknown magic class and right
	 *         is known type.
	 */
	public static boolean isSafer(Node left, Node right, Type l, Type r) {
		return (TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right)) || isTypeSafer(l, r);
	}

	/**
	 * 
	 * @param left
	 * @param right
	 * @return True if right is a magic type and left is not or if right is anonymous and left is
	 *         not.
	 */
	// public static boolean isSafer(Type left, Type right) {
	// return (!TypeFactUtil.isAnonymousMagicClass(left) && TypeFactUtil
	// .isAnonymousMagicClass(right))
	// || (!TypeFactUtil.isMagicClassKind(left) && TypeFactUtil.isMagicClassKind(right));
	// }
	/**
	 * 
	 * @param object
	 * @return True if the object is a MagicClass, a MagicPrimitive or a MagicClassArray
	 */
	public static boolean isMagicClassKind(Object object) {
		return object != null
				&& (object instanceof MagicClass || object instanceof MagicPrimitive
						|| object instanceof MagicClassArray || isMagicArray(object));
	}

	private static boolean isMagicArray(Object object) {
		boolean isMA = false;

		if (object != null && object instanceof ArrayType) {
			ArrayType at = (ArrayType) object;
			isMA = isMagicClassKind(at.base());
		}

		return isMA;
	}

	/**
	 * 
	 * @param type
	 * @return True if the type or one of its super type is a magic class kind.
	 */
	public static boolean isMagicHierarchy(Type type) {
		if (type == null) {
			return false;
		}

		Type superType = null;
		if (type instanceof ReferenceType) {
			superType = ((ReferenceType) type).superType();
		}

		return isMagicClassKind(type) || isMagicHierarchy(superType);
	}

	/**
	 * 
	 * @param node
	 * @return PPANode.isSafe or true if node is not an instance of PPANode
	 */
	public static boolean isSafe(Object node) {
		boolean isSafe = true;

		if (node != null && node instanceof PPANode) {
			PPANode sn = (PPANode) node;
			isSafe = sn.isSafe();
		}

		return isSafe;
	}

	/**
	 * 
	 * @param type
	 * @return True if the type or one of its super type is a magic class.
	 */
	public static boolean isStrictMagicHierarchy(Type type) {
		if (type == null) {
			return false;
		}

		Type superType = null;
		if (type instanceof ReferenceType) {
			superType = ((ReferenceType) type).superType();
		}

		return type instanceof MagicClass || isStrictMagicHierarchy(superType);
	}

	/**
	 * 
	 * @param object
	 * @return True if object is an instance of MagicClass with a fqn =
	 *         MagicClass.DEFAULT_MAGIC_CLASS
	 */
	public static boolean isAnonymousMagicClass(Object object) {
		boolean isAnonymous = false;

		if (object != null && object instanceof MagicClass) {
			MagicClass mc = (MagicClass) object;
			isAnonymous = mc.fullName().equals(MagicClass.DEFAULT_MAGIC_CLASS);
		}

		return isAnonymous;
	}

	/**
	 * 
	 * @param object
	 * @return True if this is a MagicClassArray[] of unknown types...
	 */
	public static boolean isAnonymousArray(Object object) {
		boolean isAnonymous = false;

		if (object instanceof ArrayType) {
			ArrayType mca = (ArrayType) object;
			isAnonymous = isAnonymousMagicClass(mca.base());
		}

		return isAnonymous;
	}

	/**
	 * 
	 * @param type1
	 * @param type2
	 * @return True if two types might be related because they have a magic class in their
	 *         hierarchy.
	 */
	public static boolean mightBeRelated(Type type1, Type type2) {
		return hasMagicInHierarchy(type1.toReference()) || hasMagicInHierarchy(type2.toReference());
	}

	public static boolean positionEquals(Position pos1, Position pos2) {
		return pos1 != null && pos2 != null
				&& pos1.nameAndLineString().equals(pos2.nameAndLineString())
				&& pos1.column() == pos2.column() && pos1.endColumn() == pos2.endColumn();
	}

	/**
	 * 
	 * @param type1
	 * @param type2
	 * @return True if
	 */
	public static boolean relatedTypes(Type type1, Type type2) {
		boolean related = false;

		if (type1 instanceof MagicClass) {
			MagicClass mType1 = (MagicClass) type1;
			related = mType1.isMagicCastValid(type2);
		} else if (type2 instanceof MagicClass) {
			MagicClass mType2 = (MagicClass) type2;
			related = mType2.isMagicCastValid(type1);
		}

		return related;
	}

	public static boolean same(Object object1, Object object2) {
		return (object1 != null && object2 != null && object1.equals(object2))
				|| (object1 == null && object2 == null);
	}

	public static boolean hasMagicSuperType(Type type) {
		boolean hasMagicST = false;

		if (type != null) {

			hasMagicST = isMagicClassKind(type.toType());
			if (!hasMagicST && type.isReference()) {
				Type parent = type.toReference().superType();
				if (parent != null) {
					hasMagicST = hasMagicSuperType(parent.toReference());
				}
			}
		}

		return hasMagicST;
	}

	public static boolean hasMagicInterface(Type type) {
		boolean hasMagicI = false;

		if (type != null) {
			hasMagicI = isMagicClassKind(type);

			for (Type interfaceT : getInterfaces(type)) {
				hasMagicI = isMagicClassKind(interfaceT);
				if (hasMagicI) {
					break;
				}
			}
		}

		return hasMagicI;
	}

	public static Set<Type> getInterfaces(Type type) {
		Set<Type> set = new HashSet<Type>();

		if (type != null && type.isReference()) {
			ReferenceType rType = type.toReference();

			for (Object obj : rType.interfaces()) {
				Type interfaceT = (Type) obj;
				set.add(interfaceT);
				set.addAll(getInterfaces(interfaceT));
			}

			set.addAll(getInterfaces(rType.superType()));
		}

		return set;
	}

	public static int getMagicType(Type type) {
		int magicType = NO_MAGIC;
		if (isAnonymousMagicClass(type)) {
			magicType = ANON_MAGIC;
		} else if (isAnonymousArray(type)) {
			magicType = ANON_MAGIC_ARRAY;
		} else if (isMagicClassKind(type)) {
			magicType = MAGIC_MAGIC;
		} else if (hasMagicSuperType(type)) {
			magicType = MAGIC_SUPER;
		} else if (hasMagicInterface(type)) {
			magicType = MAGIC_INTER;
		}
		
		return magicType;
	}

	public static boolean isTypeSafer(Type type1, Type type2) {
		return getMagicType(type1) > getMagicType(type2);
	}

}
