package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ext.jl.ast.ArrayAccess_c;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;
import soot.javaToJimple.ppa.jj.types.MagicClass;
import soot.javaToJimple.ppa.jj.types.MagicClassArray;

/**
 * <p>
 * Array access. E.g. myObject[2]
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li>Based on PPAAssignHelper.</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>the node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>array, if unsafe (e.g. myObject)</li>
 * <li>array index, if unsafe (e.g. i in myObject[i])</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Array and Array index nodes are safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPAArrayAccess_c extends ArrayAccess_c implements PPANode {

	public PPAArrayAccess_c(Position pos, Expr array, Expr index) {
		super(pos, array, index);
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		indexes.add(getMainIndex());

		if (!TypeFactUtil.isSafe(array)) {
			PPAIndex ppaIndex = TypeFactUtil.getMainIndex(array);
			if (ppaIndex != null) {
				indexes.add(ppaIndex);
			}
		}

		if (!TypeFactUtil.isSafe(index)) {
			PPAIndex ppaIndex = TypeFactUtil.getMainIndex(index);
			if (ppaIndex != null) {
				indexes.add(ppaIndex);
			}
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		// PPAIndex otherIndex = SafeUtil.getMainIndex(array);
		// TypeObject tObject = otherIndex != null ? otherIndex.getTypeObject() : null;
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(array) && TypeFactUtil.isSafe(index);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		
		try {
			// The array[i] has a new type. Change the arrayexpr type, change this node type.
			if (index.equals(getMainIndex()) && !TypeFactUtil.isSafe(array)) {
				MagicClassArray mClass = new MagicClassArray(newType.typeSystem(), newType
						.position(), newType);
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(array), array.type(), mClass,
								TypeFact.CHILD_OR_EQUALS, TypeFact.ARRAY_STRATEGY, origin));
				type = newType;
			}
			// The array has a new type!
			else if (newType.isArray()) {
				Type newArrayAccessType = newType.toArray().base();
				PPAEngine.v().reportNewFact(
						new TypeFact(getMainIndex(), type(), newArrayAccessType,
								TypeFact.CHILD_OR_EQUALS,TypeFact.ARRAY_STRATEGY, origin));
				type = newArrayAccessType;
			}
			// The index should ALWAYS be an int.
		} catch (Exception e) {
		}

		// return isSafe();
	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		TypeSystem ts = tc.typeSystem();
		Type type = null;
		if (!array.type().isArray()) {
			if (TypeFactUtil.isAnonymousMagicClass(array.type())) {
				// TODO check that it is really a MagicArray...
				MagicClass originalType = (MagicClass) array.type();
				MagicClassArray mClass = new MagicClassArray(originalType.typeSystem(),
						originalType.position(), originalType);
				if (!TypeFactUtil.isSafe(array)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(array), originalType, mClass,
									TypeFact.CHILD_OR_EQUALS,TypeFact.ARRAY_STRATEGY, null));
				}
				type = originalType;
			} else {
				throw new SemanticException("Subscript can only follow an array type.", position());
			}
		} else {
			type = array.type().toArray().base();
		}

		if (TypeFactUtil.isAnonymousMagicClass(index.type())) {
			if (!TypeFactUtil.isSafe(index)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(index), index.type(), ts.Int(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.ARRAY_STRATEGY, null));
			}
		} else if (!ts.isImplicitCastValid(index.type(), ts.Int())) {
			throw new SemanticException("Array subscript must be an integer.", position());
		}

		return type(type);
	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because this node can be used as a main index.

		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPAArrayAccess_c) {
			PPAArrayAccess_c arrayAccess = (PPAArrayAccess_c) obj;
			return TypeFactUtil.positionEquals(position, arrayAccess.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// Because this node can be used as a main index.
		// return mi.hashCode() + position.nameAndLineString().hashCode();
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}

}
