package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Binary;
import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ast.Precedence;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.jj.ast.JjBinary_c;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;
import soot.javaToJimple.ppa.jj.types.MagicClass;
import soot.javaToJimple.ppa.jj.types.MagicPrimitive;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem;

/**
 * <p>
 * Binary operation. e.g. myObj.field1 + 2;
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li></li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>this node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>left expression</li>
 * <li>right expression</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Left expression is safe</li>
 * <li>Right expression is safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPABinary_c extends JjBinary_c implements PPANode {
	public PPABinary_c(Position pos, Expr left, Binary.Operator op, Expr right) {
		super(pos, left, op, right);
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();

		indexes.add(getMainIndex());

		if (!TypeFactUtil.isSafe(left)) {
			indexes.add(TypeFactUtil.getMainIndex(left));
		}

		if (!TypeFactUtil.isSafe(right)) {
			indexes.add(TypeFactUtil.getMainIndex(right));
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(left) && TypeFactUtil.isSafe(right);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		try {
			Type pushType = null;
			if (index.equals(getMainIndex())) {
				pushType = newType;
			}
			Type returnType = inferType((PPATypeSystem) newType.typeSystem(), pushType, origin);
			if (!returnType.equalsImpl(newType) && pushType != null) {
				// Conflict!
				System.err.println("Binary Conflict!");
			}
			type = newType;
		} catch (Exception e) {
		}

	}

	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;

		try {
			node = super.typeCheck(tc);
		} catch (SemanticException se) {
			if (TypeFactUtil.isSafe(left) && TypeFactUtil.isSafe(right)) {
				throw se;
			}

		}

		if (!TypeFactUtil.isSafe(left) || !TypeFactUtil.isSafe(right)) {
			node = inferInfo(tc, null, null);
		}

		return node;
	}

	private Node inferInfo(TypeChecker tc, Type newType, TypeFact origin) throws SemanticException {
		Node node = null;
		PPATypeSystem ts = (PPATypeSystem) tc.typeSystem();

		Type returnType = inferType(ts, newType, origin);

		if (returnType.equalsImpl(ts.String()) && op == ADD) {
			node = precedence(Precedence.STRING_ADD).type(ts.String());
		} else {
			node = type(returnType);
		}

		return node;
	}

	// TODO: ADD INFERED TYPE (like return type of calls...)
	private Type inferType(PPATypeSystem ts, Type newType, TypeFact origin) throws SemanticException {
		Type returnType = null;
		MagicClass magicClass = ts.magic();
		MagicPrimitive intPrimitive = ts.magicInt();
		// MagicPrimitive booleanPrimitive = ts.magicBoolean();
		Type l = left.type();
		Type r = right.type();
		boolean isBoolean = newType != null && newType.isBoolean();
		boolean isNumeric = newType != null && newType.isNumeric();
		boolean isString = newType != null && newType.equalsImpl(ts.String());

		if (op == GT || op == LT || op == GE || op == LE) {
			if (TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right) && l.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, l,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			} else if (TypeFactUtil.isSafe(right) && !TypeFactUtil.isSafe(left) && r.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, r,
								TypeFact.SUPER_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			} else if (!TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, intPrimitive,
								TypeFact.SUPER_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			} else {
				throw new SemanticException("The " + op + " operator must have numeric operands.",
						left.position());
			}
			returnType = ts.Boolean();
		} else if (op == EQ || op == NE) {
			if (TypeFactUtil.isSafer(left, right, l, r)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, l,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			} else if (TypeFactUtil.isSafer(right, left, r, l)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, r,
								TypeFact.SUPER_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			} else {
				// TODO Should report related!
			}
			returnType = ts.Boolean();
		} else if (op == COND_OR || op == COND_AND) {
			if (!l.isBoolean() && !TypeFactUtil.isSafe(left)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.Boolean(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			}
			if (!r.isBoolean() && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.Boolean(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
			}
			returnType = ts.Boolean();
		} else if (op == ADD) {
			if (ts.equals(l, ts.String()) || ts.equals(r, ts.String())) {
				// We cannot do any other inference.
				returnType = ts.String();
			} else {
				// For the first two: since no expression is a string, but one of the expression is
				// an object, the other must be a string!
				if (!TypeFactUtil.isAnonymousMagicClass(l) && (!l.isNumeric() || isString)
						&& TypeFactUtil.isAnonymousMagicClass(r) && !TypeFactUtil.isSafe(right)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.String(),
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
					returnType = ts.String();
				}
				if (!TypeFactUtil.isAnonymousMagicClass(r) && (!r.isNumeric() || isString)
						&& TypeFactUtil.isAnonymousMagicClass(l) && !TypeFactUtil.isSafe(l)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.String(),
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
					returnType = ts.String();
				}
				if (isNumeric) {
					if (TypeFactUtil.isAnonymousMagicClass(l) && !TypeFactUtil.isSafe(left)) {
						PPAEngine.v().reportNewFact(
								new TypeFact(TypeFactUtil.getMainIndex(left), l, intPrimitive,
										TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
					}
					if (TypeFactUtil.isAnonymousMagicClass(r) && !TypeFactUtil.isSafe(right)) {
						PPAEngine.v().reportNewFact(
								new TypeFact(TypeFactUtil.getMainIndex(right), r, intPrimitive,
										TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
					}
					returnType = newType;
				}

				if (returnType == null) {
					// Catch all.
					// throw new SemanticException();
					returnType = magicClass.toType();
				}

			}
		} else if (op == SUB || op == MUL || op == DIV || op == MOD) {
			if (TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right) && l.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, l,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = l;
			} else if (TypeFactUtil.isSafe(right) && !TypeFactUtil.isSafe(left) && r.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, r,
								TypeFact.SUPER_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = r;
			} else if (TypeFactUtil.isAnonymousMagicClass(l)
					&& TypeFactUtil.isAnonymousMagicClass(r)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.Int(),
								TypeFact.SUPER_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.Int(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = ts.Int();
			} else {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, intPrimitive,
								TypeFact.UNKNOWN,TypeFact.BINARY_STRATEGY, origin));
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, intPrimitive,
								TypeFact.UNKNOWN,TypeFact.BINARY_STRATEGY, origin));
				returnType = intPrimitive.toType();
				// throw new SemanticException("The " + op + " operator must have numeric
				// operands.",
				// left.position());
			}
		} else if (op == BIT_AND || op == BIT_OR || op == BIT_XOR) {
			if (isBoolean) {
				if (!TypeFactUtil.isSafe(left)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.Boolean(),
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				}
				if (!TypeFactUtil.isSafe(right)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.Boolean(),
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				}
				returnType = ts.Boolean();
			} else if (isNumeric) {
				if (!TypeFactUtil.isSafe(left)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(left), l, intPrimitive,
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				}
				if (!TypeFactUtil.isSafe(right)) {
					PPAEngine.v().reportNewFact(
							new TypeFact(TypeFactUtil.getMainIndex(right), r, intPrimitive,
									TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				}
				returnType = newType;
			} else if (l.isBoolean() && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.Boolean(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = ts.Boolean();
			} else if (r.isBoolean() && !TypeFactUtil.isSafe(left)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.Boolean(),
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = ts.Boolean();
			} else if (l.isNumeric() && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, l,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = l;
			} else if (r.isNumeric() && !TypeFactUtil.isSafe(left)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, r,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = r;
			} else if (!TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right)) {
				// PPAEngine.v().reportNewFact(
				// new TypeFact(TypeFactUtil.getMainIndex(left), l, ts.Int(),
				// TypeFact.CHILD_OR_EQUALS));
				// PPAEngine.v().reportNewFact(
				// new TypeFact(TypeFactUtil.getMainIndex(right), r, ts.Int(),
				// TypeFact.CHILD_OR_EQUALS));
				returnType = magicClass.toType();
			}
		} else if (op == SHL || op == SHR || op == USHR) {
			if (l.isNumeric() && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, l,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = l;
			} else if (r.isNumeric() && !TypeFactUtil.isSafe(left)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, r,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = r;
			} else if (!TypeFactUtil.isSafe(left) && !TypeFactUtil.isSafe(right)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(left), l, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(right), r, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.BINARY_STRATEGY, origin));
				returnType = intPrimitive.toType();
			}
		}

		return returnType;
	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPABinary_c) {
			PPABinary_c ppaObj = (PPABinary_c) obj;
			return TypeFactUtil.positionEquals(position, ppaObj.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}
}
