package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import polyglot.ast.Call;
import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ast.Receiver;
import polyglot.ast.Special;
import polyglot.ast.TypeNode;
import polyglot.ext.jl.ast.Call_c;
import polyglot.ext.jl.types.MethodInstance_c;
import polyglot.types.ClassType;
import polyglot.types.Context;
import polyglot.types.Flags;
import polyglot.types.MethodInstance;
import polyglot.types.ProcedureInstance;
import polyglot.types.ReferenceType;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;
import soot.javaToJimple.ppa.jj.types.MagicClass;
import soot.javaToJimple.ppa.jj.types.MagicMethodInstance_c;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem;

/**
 * <p>
 * Call. e.g. method(...);
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li></li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>this node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>arguments (parameters)</li>
 * <li>target</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>See javadoc.</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPACall_c extends Call_c implements PPANode {

	private ClassType currentClass;

	private Type fixedReturnType;
	
	public static int maxCountArgs = 0;
	
	public PPACall_c(Position pos, Receiver target, String name, List arguments) {
		super(pos, target, name, arguments);
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();
		indexes.add(getMainIndex());

		for (Object arg : arguments) {
			if (!TypeFactUtil.isSafe(arg)) {
				indexes.add(TypeFactUtil.getMainIndex(arg));
			}
		}

		if (!TypeFactUtil.isSafe(target)) {
			indexes.add(TypeFactUtil.getMainIndex(target));
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	/**
	 * <p>
	 * A call is unsafe if:
	 * <ul>
	 * <li>It's return type is an unknown class.</li>
	 * <li>It's container is an unknown class!</li>
	 * <li>Or if there is more than one possible methods according to its parameter types.</li>
	 * </ul>
	 * </p>
	 * 
	 * @return True if safe. False if not.
	 */
	public boolean isSafe() {
		boolean isSafe = !TypeFactUtil.isMagicClassKind(mi.returnType())
				&& !TypeFactUtil.isMagicHierarchy(mi.container());

		if (isSafe) {
			List<Type> argTypes = getArgTypes();

			try {
				ReferenceType targetType = this.findTargetType();
				PPATypeSystem ppaTs = (PPATypeSystem) targetType.typeSystem();
				List<MethodInstance> mis = ppaTs.findAcceptableMethods(targetType, this.name,
						argTypes, currentClass);

				isSafe = !TypeFactUtil.containsAnonymous(argTypes) || mis.size() == 1;

				if (mis.size() > 0) {
					ProcedureInstance piTemp = ppaTs.findProcedure(mis, targetType, argTypes,
							currentClass);
					isSafe = isSafe && piTemp != null;
				}
			} catch (SemanticException se) {
				se.printStackTrace();
			}
		}
		return isSafe;
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		List<Type> argTypes = getArgTypes();
		ReferenceType targetType = null;
		PPATypeSystem ts = null;
		
		// Change main index return type!
		// We only compare the call position, because mi might have changed
		// since the node was declared unsafe.
		Object addIndex = index.getAdditionalIndex();
		if (addIndex != null && addIndex.equals(this)) {
			try {
				targetType = findTargetType();
				ts = (PPATypeSystem) targetType.typeSystem();
				MethodInstance mi = typeCheckMethodInstance(ts, targetType, this.name,
						currentClass, argTypes, newType);

				// XXX It's possible that we did not find any new methods
				if (mi != null) {
					this.mi = mi;
					this.type = newType;
					// this.type = mi.returnType();

					this.fixedReturnType = this.type;

					// Only report safer arguments if we are SURE that those are the real argument
					// types!
					List<MethodInstance> mis = ts.findAcceptableMethods(targetType, this.name,
							argTypes, currentClass);
					if (ts.getAcceptable(mis, newType).size() == 1) {
						reportSaferArguments(mi, origin);
					}
				}
			} catch (SemanticException se) {
				se.printStackTrace();
			}
		}
		// One of the arguments changed or it was the target. Do type check again!
		else {
			try {
				targetType = findTargetType();
				ts = (PPATypeSystem) targetType.typeSystem();
				MethodInstance mi = null;

				if (fixedReturnType != null) {
					mi = typeCheckMethodInstance(ts, targetType, this.name, currentClass, argTypes,
							fixedReturnType);
				} else {
					mi = typeCheckMethodInstance(ts, targetType, this.name, currentClass, argTypes);
				}

				// XXX No need to report if arguments were made safer, because anyway, we will
				// report this if the return type changed.
				// XXX We use old mi, because all other nodes refer to this old mi and not the new
				// one!
				// XXX It is possible that we cannot find a method that matches this argument type
				// (e.g.): because the argument is an inferred super class and the type system won't
				// find a valid method).
				if (mi != null) {
					if (!ts.equals(this.type, mi.returnType())) {
						PPAEngine.v().reportNewFact(
								new TypeFact(TypeFactUtil.getMainIndex(this), this.type, mi
										.returnType(), TypeFact.UNKNOWN,TypeFact.METHOD_STRATEGY, origin));
					}
					this.type = mi.returnType();
					this.mi = mi;
				}
				// this.type = newType;
			} catch (SemanticException se) {
				se.printStackTrace();
			}
		}

		// return isSafe();
	}

	private void reportSaferArguments(MethodInstance mi, TypeFact origin) {
		int countArgs = 0;
		int size = mi.formalTypes().size();
		for (int i = 0; i < size; i++) {
			Type miType = (Type) mi.formalTypes().get(i);
			Expr expr = (Expr) arguments.get(i);
			Type exprType = expr.type();
			// Only report a safer argument if it is actually safer!
			if (!TypeFactUtil.isSafe(expr) && !expr.type().equalsImpl(miType)
					&& TypeFactUtil.isTypeSafer(miType, exprType)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(expr), exprType, miType,
								TypeFact.CHILD_OR_EQUALS,TypeFact.METHOD_STRATEGY, origin));
				countArgs++;
			}
		}
		
		if (countArgs > maxCountArgs) {
			maxCountArgs = countArgs;
		}
	}

	@Override
	public ReferenceType findTargetType() throws SemanticException {
		ReferenceType t = null;

		try {
			t = super.findTargetType();
		} catch (SemanticException se) {
			t = (ReferenceType) PPAEngine.v().getTypeSystem().forName(
					MagicClass.DEFAULT_MAGIC_CLASS);
		}
		return t;
	}

	public void pushSaferArguments() {
		try {
			ReferenceType targetType = this.findTargetType();
			PPATypeSystem ppaTs = (PPATypeSystem) targetType.typeSystem();
			List<MethodInstance> mis = ppaTs.findAcceptableMethods(targetType, this.name,
					getArgTypes(), currentClass);

			// Only push info if we are SURE that this is the only method!
			if (mis.size() == 1 || ppaTs.getAcceptable(mis, fixedReturnType).size() == 1) {
				int size = mi.formalTypes().size();
				for (int i = 0; i < size; i++) {
					Type miType = (Type) mi.formalTypes().get(i);
					Expr expr = (Expr) arguments.get(i);
					if (!TypeFactUtil.isSafe(expr)
							&& TypeFactUtil.isAnonymousMagicClass(expr.type())
							&& !expr.type().equalsImpl(miType)) {
						PPAEngine.v().reportNewFact(
								new TypeFact(TypeFactUtil.getMainIndex(expr), expr.type(), miType,
										TypeFact.CHILD_OR_EQUALS,TypeFact.METHOD_STRATEGY, null));
					}
				}
			}
		} catch (SemanticException se) {
			se.printStackTrace();
		}
	}

	public MethodInstance typeCheckMethodInstance(TypeSystem ts, ReferenceType targetType,
			String name, ClassType currentClass, List argTypes) throws SemanticException {

		MethodInstance mi = ts.findMethod(targetType, this.name, argTypes, currentClass);
		staticCheck(mi, targetType);
		return mi;
	}

	public MethodInstance typeCheckMethodInstance(PPATypeSystem ts, ReferenceType targetType,
			String name, ClassType currentClass, List argTypes, Type expectedReturnType)
			throws SemanticException {

		MethodInstance mi = ts.findMethod(targetType, this.name, argTypes, currentClass,
				expectedReturnType);
		staticCheck(mi, targetType);
		return mi;
	}

	/**
	 * This call is in a static context if and only if the target (possibly implicit) is a type
	 * node.
	 */
	private void staticCheck(MethodInstance mi, ReferenceType targetType) throws SemanticException {

		boolean staticContext = (this.target instanceof TypeNode);

		if (staticContext && !mi.flags().isStatic()) {
			if (mi.container() instanceof MagicClass) {
				// XXX we should log somewhere that the method is now static
				// instead of hacking through polyglot...
				if (mi instanceof MethodInstance_c) {
					MagicMethodInstance_c newMic = (MagicMethodInstance_c) mi;
					newMic.setFlags(mi.flags().set(Flags.STATIC));
				}
			} else {
				// throw new SemanticException("Cannot call non-static method " + this.name + " of "
				// + targetType + " in static " + "context.", this.position());
			}
		}

		// If the target is super, but the method is abstract, then complain.
		if (this.target instanceof Special && ((Special) this.target).kind() == Special.SUPER
				&& mi.flags().isAbstract()) {
			// XXX Don't throw exception just to bother the poor Bart...
			// throw new SemanticException("Cannot call an abstract method " + "of the super class",
			// this.position());
		}
	}

	/** Type check the call. */
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Call node = null;
		PPATypeSystem ts = (PPATypeSystem) tc.typeSystem();
		Context c = tc.context();
		currentClass = c.currentClass();

		List<Type> argTypes = getArgTypes();

		if (this.target == null) {
			node = (Call) this.typeCheckNullTarget(tc, argTypes);
		} else {

			ReferenceType targetType = this.findTargetType();

			MethodInstance mi = typeCheckMethodInstance(ts, targetType, this.name, currentClass,
					argTypes);

			// If we found a method, the call must type check, so no need to check
			// the arguments here.

			Type returnType = null;
			try {
				returnType = mi.returnType();
			} catch (Exception e) {
				e.printStackTrace();
			}

			// If the method is not safe, it does not make sense to force a specific return type
			// somewhere.
			if (!isSafe()) {
				returnType = ts.typeForName(MagicClass.DEFAULT_MAGIC_CLASS);
			}

			PPACall_c call = (PPACall_c) this.methodInstance(mi).type(returnType);
			call.checkConsistency(c);
			node = call;
		}

		// reportSaferArguments(node.methodInstance());

		return node;
	}

	private List<Type> getArgTypes() {
		List<Type> argTypes = new ArrayList<Type>(this.arguments.size());

		for (Iterator i = this.arguments.iterator(); i.hasNext();) {
			Expr e = (Expr) i.next();
			argTypes.add(e.type());
		}

		return argTypes;
	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPACall_c) {
			PPACall_c call = (PPACall_c) obj;
			return TypeFactUtil.positionEquals(position, call.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}

}
