package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ext.jl.ast.Conditional_c;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;

/**
 * <p>
 * Conditional. e.g. test ? doThat() : doThis();
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li></li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>this node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>consequent</li>
 * <li>alternative</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Condition is safe</li>
 * <li>Consequent is safe</li>
 * <li>Alternative is safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPAConditional_c extends Conditional_c implements PPANode {
	public PPAConditional_c(Position pos, Expr cond, Expr consequent, Expr alternative) {
		super(pos, cond, consequent, alternative);
	}

	public List<PPAIndex> getIndexes() {
		// TODO Should we include condition?
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();

		indexes.add(getMainIndex());

		if (!TypeFactUtil.isSafe(consequent)) {
			indexes.add(TypeFactUtil.getMainIndex(consequent));
		}

		if (!TypeFactUtil.isSafe(alternative)) {
			indexes.add(TypeFactUtil.getMainIndex(alternative));
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(cond) && TypeFactUtil.isSafe(consequent)
				&& TypeFactUtil.isSafe(alternative);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		try {
			Type pushType = null;
			if (index.equals(getMainIndex())) {
				pushType = newType;
			}
			inferType(pushType, origin);
			type = newType;
		} catch (Exception e) {
		}

	}

	@Override
	/** Type check the expression. */
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;
		try {
			node = super.typeCheck(tc);
		} catch (SemanticException se) {
			if (isSafe()) {
				throw se;
			}
		}

		// Check condition
		Type condType = cond.type();
		if (TypeFactUtil.isAnonymousMagicClass(condType) && !TypeFactUtil.isSafe(cond)) {
			PPAEngine.v().reportNewFact(
					new TypeFact(TypeFactUtil.getMainIndex(cond), condType, tc.typeSystem()
							.Boolean(), TypeFact.CHILD_OR_EQUALS,TypeFact.CONDITION_STRATEGY, null));
		}

		// Check consequent and alternative
		Type returnType = inferType(null, null);
		if (returnType != null) {
			node = type(returnType);
		} else {
			// Should not happen...
			// throw se;
			node = this;
		}

		return node;
	}

	private Type inferType(Type newType, TypeFact origin) {
		Type returnType = null;
		Type t1 = consequent.type();
		Type t2 = alternative.type();

		if (newType != null) {
			if (!TypeFactUtil.isSafe(consequent)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(consequent), t1, newType,
								TypeFact.UNKNOWN,TypeFact.CONDITIONAL_STRATEGY,origin));
			}
			if (!TypeFactUtil.isSafe(alternative)) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(alternative), t2, newType,
								TypeFact.UNKNOWN,TypeFact.CONDITIONAL_STRATEGY,origin));
			}
			returnType = newType;
		} else if (t1.equalsImpl(t2)) {
			return t1;
		} else if (t1.isNumeric() && TypeFactUtil.isAnonymousMagicClass(t2)
				&& !TypeFactUtil.isSafe(alternative)) {
			PPAEngine.v().reportNewFact(
					new TypeFact(TypeFactUtil.getMainIndex(alternative), t2, t1,
							TypeFact.CHILD_OR_EQUALS, TypeFact.CONDITIONAL_STRATEGY, origin));
			returnType = t1;
		} else if (t2.isNumeric() && TypeFactUtil.isAnonymousMagicClass(t1)
				&& !TypeFactUtil.isSafe(consequent)) {
			PPAEngine.v().reportNewFact(
					new TypeFact(TypeFactUtil.getMainIndex(consequent), t1, t2,
							TypeFact.CHILD_OR_EQUALS,TypeFact.CONDITIONAL_STRATEGY,origin));
			returnType = t2;
		} else if (TypeFactUtil.isTypeSafer(t1, t2) && !TypeFactUtil.isSafe(alternative)) {
			PPAEngine.v().reportNewFact(
					new TypeFact(TypeFactUtil.getMainIndex(alternative), t2, t1,
							TypeFact.CHILD_OR_EQUALS,TypeFact.CONDITIONAL_STRATEGY,origin));
			returnType = t1;
		} else if (TypeFactUtil.isTypeSafer(t2, t1) && !TypeFactUtil.isSafe(consequent)) {
			PPAEngine.v().reportNewFact(
					new TypeFact(TypeFactUtil.getMainIndex(consequent), t1, t2,
							TypeFact.CHILD_OR_EQUALS,TypeFact.CONDITIONAL_STRATEGY,origin));
			returnType = t2;
		} else if (TypeFactUtil.hasMagicInHierarchy(t1.toReference())
				&& TypeFactUtil.hasMagicInHierarchy(t2.toReference())) {
			// Arbitrarily, we choose t1 as the type of this node...
			returnType = t1;
		} else {
			// Catch all!
			returnType = t1;
		}

		return returnType;
	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPAConditional_c) {
			PPAConditional_c conditional = (PPAConditional_c) obj;
			return TypeFactUtil.positionEquals(position, conditional.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}
}
