package soot.javaToJimple.ppa.jj.ast;

import java.util.ArrayList;
import java.util.List;

import polyglot.ast.Expr;
import polyglot.ast.Node;
import polyglot.ast.Unary;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.Position;
import polyglot.visit.TypeChecker;
import soot.javaToJimple.jj.ast.JjUnary_c;
import soot.javaToJimple.ppa.PPAEngine;
import soot.javaToJimple.ppa.PPAIndex;
import soot.javaToJimple.ppa.PPANode;
import soot.javaToJimple.ppa.TypeFact;
import soot.javaToJimple.ppa.TypeFactUtil;
import soot.javaToJimple.ppa.jj.types.MagicPrimitive;
import soot.javaToJimple.ppa.jj.types.PPATypeSystem;

/**
 * <p>
 * Unary operation. e.g. myObj.field1++;
 * </p>
 * 
 * <p>
 * <strong>Inference strategies</strong>
 * <ul>
 * <li></li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Main index:</strong>
 * <ul>
 * <li>this node</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Indexes:</strong>
 * <ul>
 * <li>main index</li>
 * <li>expression</li>
 * </ul>
 * </p>
 * 
 * <p>
 * <strong>Safe condition:</strong>
 * <ul>
 * <li>Expression is safe</li>
 * </ul>
 * </p>
 * 
 * @author barthelemy
 * 
 */
public class PPAUnary_c extends JjUnary_c implements PPANode {
	public PPAUnary_c(Position pos, Unary.Operator op, Expr expr) {
		super(pos, op, expr);
	}

	public List<PPAIndex> getIndexes() {
		List<PPAIndex> indexes = new ArrayList<PPAIndex>();

		indexes.add(getMainIndex());

		if (!TypeFactUtil.isSafe(expr)) {
			indexes.add(TypeFactUtil.getMainIndex(expr));
		}

		return indexes;
	}

	public PPAIndex getMainIndex() {
		return new PPAIndex(null, this);
	}

	public boolean isSafe() {
		return TypeFactUtil.isSafe(expr);
	}

	public void makeSafe(PPAIndex index, TypeFact origin) {
		Type newType = origin.getNewType();
		try {
			// The new type should NEVER be different (except if we go into the details of numerical
			// hierarchy), because the operation determines the type.
			inferType((PPATypeSystem) newType.typeSystem(), origin);
			type = newType;
		} catch (Exception e) {
		}

	}

	@Override
	public Node typeCheck(TypeChecker tc) throws SemanticException {
		Node node = null;
		Node tempNode = null;
		SemanticException sExcep = null;
		try {
			node = super.typeCheck(tc);
		} catch (SemanticException se) {
			sExcep = se;
		}

		if (sExcep != null && TypeFactUtil.isSafe(expr)) {
			// throw sExcep;
		} else if (!TypeFactUtil.isSafe(expr)){
			// InferInfo does inference + produces a node
			tempNode = inferInfo(tc, null);
			if (node == null) {
				node = tempNode;
			}
		}

		return node;
	}

	private Node inferInfo(TypeChecker tc, TypeFact origin) {

		Type returnType = inferType((PPATypeSystem) tc.typeSystem(), origin);
		Node returnNode = null;

		if (returnType != null) {
			returnNode = type(returnType);
		} else {
			returnNode = this;
		}

		return returnNode;
	}

	private Type inferType(PPATypeSystem ts, TypeFact origin) {
		Type oldType = expr.type();
		Type returnType = null;
		MagicPrimitive intPrimitive = ts.magicInt();
		MagicPrimitive booleanPrimitive = ts.magicBoolean();

		if (op == POST_INC || op == POST_DEC || op == PRE_INC || op == PRE_DEC) {

			if (!oldType.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(expr), oldType, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.UNARY_STRATEGY,origin));
			}

			// if (!(expr instanceof Variable)) {
			// throw new SemanticException("Operand of " + op + " operator must be a variable.",
			// expr.position());
			// }
			//
			// if (((Variable) expr).flags().isFinal()) {
			// throw new SemanticException("Operand of " + op
			// + " operator must be a non-final variable.", expr.position());
			// }

			returnType = intPrimitive.toType();
		}

		if (op == BIT_NOT) {
			if (TypeFactUtil.isAnonymousMagicClass(oldType)
					|| !ts.isImplicitCastValid(oldType, ts.Long())) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(expr), oldType, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.UNARY_STRATEGY,origin));
			}

			returnType = intPrimitive.toType();
		}

		if (op == NEG || op == POS) {
			if (!oldType.isNumeric()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(expr), oldType, intPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.UNARY_STRATEGY,origin));
			}

			returnType = intPrimitive.toType();
		}

		if (op == NOT) {
			if (!oldType.isBoolean()) {
				PPAEngine.v().reportNewFact(
						new TypeFact(TypeFactUtil.getMainIndex(expr), oldType, booleanPrimitive,
								TypeFact.CHILD_OR_EQUALS,TypeFact.UNARY_STRATEGY,origin));
			}

			returnType = booleanPrimitive.toType();
		}

		return returnType;
	}

	@Override
	public boolean equals(Object obj) {
		// XXX Because mi might have changed, but the call is at the same place.
		if (obj instanceof PPAUnary_c) {
			PPAUnary_c ppaObj = (PPAUnary_c) obj;
			return TypeFactUtil.positionEquals(position, ppaObj.position);
		}
		return false;
	}

	@Override
	public int hashCode() {
		// XXX Because mi might have changed, but the call is at the same place.
		return TypeFactUtil.hashCode(position);
	}
}
