/* Soot - a J*va Optimization Framework
 * Copyright (C) 2004 Jennifer Lhotak
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package soot.javaToJimple.ppa.jj.types;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import polyglot.ext.jl.types.ConstructorInstance_c;
import polyglot.ext.jl.types.FieldInstance_c;
import polyglot.frontend.ExtensionInfo;
import polyglot.main.Report;
import polyglot.types.CachingResolver;
import polyglot.types.ClassType;
import polyglot.types.ConstructorInstance;
import polyglot.types.FieldInstance;
import polyglot.types.Flags;
import polyglot.types.MemberInstance;
import polyglot.types.MethodInstance;
import polyglot.types.NoMemberException;
import polyglot.types.ParsedClassType;
import polyglot.types.ProcedureInstance;
import polyglot.types.ReferenceType;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.util.InternalCompilerError;
import polyglot.util.Position;
import soot.javaToJimple.jj.types.JjTypeSystem_c;
import soot.javaToJimple.ppa.TypeFactUtil;

public class PPATypeSystem_c extends JjTypeSystem_c implements PPATypeSystem {

	protected final MagicPrimitive MAGIC_INT = new MagicPrimitive(INT_);

	protected final MagicPrimitive MAGIC_BOOLEAN = new MagicPrimitive(BOOLEAN_);

	/**
	 * <p>
	 * Check if all fields belong to a magic container
	 * </p>
	 * 
	 * @param fields
	 * @return
	 */
	private boolean allFromMagicContainer(Collection fields) {
		boolean allMagicContainer = true;
		for (Object fiObject : fields) {
			FieldInstance fiTemp = (FieldInstance) fiObject;
			if (!TypeFactUtil.isMagicHierarchy((fiTemp.container()))) {
				allMagicContainer = false;
				break;
			}
		}

		return allMagicContainer && fields.size() > 0;
	}

	@Override
	public boolean canOverride(MethodInstance mi, MethodInstance mj) {
		boolean canOver = super.canOverride(mi, mj);

		if (!canOver && (TypeFactUtil.hasMagic(mi) || TypeFactUtil.hasMagic(mj))) {
			// TODO ADD INFERENCE HERE!!!
			canOver = true;
		}

		return canOver;
	}

	/**
	 * <p>
	 * This method is a copy of the super method with some modifications at the beginning and at the
	 * end.
	 * </p>
	 */
	@Override
	public void checkClassConformance(ClassType ct) throws SemanticException {
		boolean isMagic = ct instanceof MagicClass;
		boolean isMagicHierarchy = TypeFactUtil.hasMagicInHierarchy(ct);

		if (ct.flags().isAbstract() || (!isMagic && isMagicHierarchy)) {
			// don't need to check abstract classes and interfaces
			return;
		}

		// build up a list of superclasses and interfaces that ct
		// extends/implements that may contain abstract methods that
		// ct must define.
		List superInterfaces = abstractSuperInterfaces(ct);

		// check each abstract method of the classes and interfaces in
		// superInterfaces
		for (Iterator i = superInterfaces.iterator(); i.hasNext();) {
			ReferenceType rt = (ReferenceType) i.next();
			for (Iterator j = rt.methods().iterator(); j.hasNext();) {
				MethodInstance mi = (MethodInstance) j.next();
				if (!mi.flags().isAbstract()) {
					// the method isn't abstract, so ct doesn't have to
					// implement it.
					continue;
				}

				boolean implFound = false;
				ReferenceType curr = ct;
				while (curr != null && !implFound) {
					List possible = curr.methods(mi.name(), mi.formalTypes());
					for (Iterator k = possible.iterator(); k.hasNext();) {
						MethodInstance mj = (MethodInstance) k.next();
						if (!mj.flags().isAbstract()
								&& ((isAccessible(mi, ct) && isAccessible(mj, ct)) || isAccessible(
										mi, mj.container().toClass()))) {
							// The method mj may be a suitable implementation of mi.
							// mj is not abstract, and either mj's container
							// can access mi (thus mj can really override mi), or
							// mi and mj are both accessible from ct (e.g.,
							// mi is declared in an interface that ct implements,
							// and mj is defined in a superclass of ct).

							// If neither the method instance mj nor the method
							// instance mi is declared in the class type ct, then
							// we need to check that it has appropriate protections.
							if (!equals(ct, mj.container()) && !equals(ct, mi.container())) {
								try {
									// check that mj can override mi, which
									// includes access protection checks.
									checkOverride(mj, mi);
								} catch (SemanticException e) {
									// change the position of the semantic
									// exception to be the class that we
									// are checking.
									throw new SemanticException(e.getMessage(), ct.position());
								}
							} else {
								// the method implementation mj or mi was
								// declared in ct. So other checks will take
								// care of access issues
							}
							implFound = true;
							break;
						}
					}

					if (curr == mi.container()) {
						// we've reached the definition of the abstract
						// method. We don't want to look higher in the
						// hierarchy; this is not an optimization, but is
						// required for correctness.
						break;
					}

					curr = curr.superType() == null ? null : curr.superType().toReference();
				}

				// did we find a suitable implementation of the method mi?
				if (!implFound && !ct.flags().isAbstract()) {
					if (isMagic) {
						MethodInstance mi2 = new MagicMethodInstance_c(mi.typeSystem(), mi
								.position(), ct, mi.flags(), mi.returnType(), mi.name(), mi
								.formalTypes(), mi.throwTypes());
						MagicClass mc = (MagicClass) ct;
						mc.getMethods().add(mi2);
					} else {
						throw new SemanticException(ct.fullName() + " should be "
								+ "declared abstract; it does not define " + mi.signature()
								+ ", which is declared in " + rt.toClass().fullName(), ct
								.position());
					}
				}
			}
		}
	}

	@Override
	public void checkOverride(MethodInstance mi, MethodInstance mj) throws SemanticException {
		try {
			super.checkOverride(mi, mj);
		} catch (SemanticException se) {
			// TODO ADD INFERENCE HERE!!!
			if (!TypeFactUtil.hasMagic(mi) && !TypeFactUtil.hasMagic(mj)) {
				throw se;
			}
		}
	}

	/**
	 * <p>
	 * This methods clean an argument list from anonymous magic type by changing the anonymous type
	 * by its super type if available.
	 * <p>
	 * 
	 * <p>
	 * For example: call(new MyMagicClass() {...},myInt) would be changed to
	 * call(MyMagicClass,myInt).
	 * <p>
	 * 
	 * @param args
	 * @return
	 */
	private List cleanArgsFromAnon(List args) {
		List newArgs = new ArrayList();

		for (Object arg : args) {
			if (arg instanceof ClassType) {
				ClassType classArg = (ClassType) arg;
				Type superType = classArg.superType();
				if (classArg.isAnonymous() && superType != null && superType instanceof MagicClass) {
					newArgs.add(superType);
				} else {
					newArgs.add(arg);
				}
			} else {
				newArgs.add(arg);
			}
		}

		return newArgs;
	}

	/**
	 * 
	 * @param argTypes
	 * @return True if one of the argument types is a magic class.
	 */
	private boolean containsMagicClass(List argTypes) {
		boolean containsMagicClass = false;

		for (Object temp : argTypes) {
			Type type = (Type) temp;
			if (type instanceof MagicClass) {
				containsMagicClass = true;
				break;
			}
		}

		return containsMagicClass;
	}

	@Override
	public List<MethodInstance> findAcceptableMethods(ReferenceType container, String name,
			List argTypes, ClassType currClass) throws SemanticException {
		List<MethodInstance> acceptables = super.findAcceptableMethods(container, name, argTypes,
				currClass);

		return refineAcceptables(acceptables, argTypes);
	}

	@Override
	public ConstructorInstance findConstructor(ClassType container, List argTypes,
			ClassType currClass) throws SemanticException {
		ConstructorInstance ci = null;
		assert_(container);
		assert_(argTypes);
		List acceptable = null;

		acceptable = super.findAcceptableConstructors(container, argTypes, currClass);

		if (acceptable.size() == 0) {
			if (container instanceof MagicClass) {
				ci = new ConstructorInstance_c(container.typeSystem(),
						Position.compilerGenerated(), container, Flags.PUBLIC,
						cleanArgsFromAnon(argTypes), new ArrayList());
				MagicClass magicClass = (MagicClass) container;
				magicClass.getConstructors().add(ci);
				acceptable.add(ci);
			} else if (container instanceof ParsedClassType) {
				MagicClass magicClass = (MagicClass)container.typeSystem().forName(MagicClass.MAGIC_PACKAGE + "." + container.name());
				ci = new ConstructorInstance_c(magicClass.typeSystem(),
						Position.compilerGenerated(), magicClass, Flags.PUBLIC,
						cleanArgsFromAnon(argTypes), new ArrayList());
				magicClass.getConstructors().add(ci);
				acceptable.add(ci);
			}
			else {
				throw new NoMemberException(NoMemberException.CONSTRUCTOR,
						"No valid constructor found for " + container + "("
								+ listToString(argTypes) + ").");
			}
		}

		ci = (ConstructorInstance) super.findProcedure(acceptable, container, argTypes, currClass);

		if (ci == null) {
			/*
			 * Unfortunately, this can happen because we generate missing constructors.
			 * 
			 * E.g.: new MyClass(magicType1); new MyClass(magicType2); new MyClass(object).
			 * 
			 * In this case, we would generate two constructors accepting one parameter. When
			 * encountering MyClass(object), we would have two possible constructors. It means that,
			 * if the code is compilable, there is really only one defined constructor with one
			 * parameter, but it is not possible to know which one.
			 * 
			 * By default, we call the first acceptable constructor.
			 */
			ci = (ConstructorInstance) acceptable.get(0);
		}

		return ci;
	}

	public FieldInstance findField(ReferenceType container, java.lang.String name,
			ClassType currClass, boolean finalRequest) throws SemanticException {
		
		Collection fields = findFields(container, name);
		FieldInstance fi = null;

		if (fields.size() == 0) {
			// Field from a magic class.
			if (finalRequest && container instanceof MagicClass) {
				FieldInstance temp = new FieldInstance_c(container.typeSystem(), Position
						.compilerGenerated(), container, Flags.PUBLIC, container.typeSystem()
						.typeForName(MagicClass.DEFAULT_MAGIC_CLASS), name);
				MagicClass magicClass = (MagicClass) container;
				magicClass.getFields().add(temp);
				fi = temp;
			}
			// Field from a class with a magic class in its hierarchy!
			else if (finalRequest && TypeFactUtil.isStrictMagicHierarchy(container)) {
				MagicClass type = TypeFactUtil.getFirstMagicClassInHierarchy(container);
				FieldInstance temp = new FieldInstance_c(type.typeSystem(), Position
						.compilerGenerated(), type.toReference(), Flags.PUBLIC, container
						.typeSystem().typeForName(MagicClass.DEFAULT_MAGIC_CLASS), name);
				type.getFields().add(temp);
				fi = temp;
			}
			// Container is not a magic type AND super type is not a magic type... Maybe field is in
			// magic interfaces?
			else if (finalRequest && container instanceof ParsedClassType
					&& TypeFactUtil.hasMagicInHierarchy(container)) {
				// Because there might be more than one interface, put the field in the container
				// (would be a constant anyway).
				FieldInstance temp = new FieldInstance_c(container.typeSystem(), Position
						.compilerGenerated(), container, Flags.PUBLIC.set(Flags.FINAL).set(
						Flags.STATIC), container.typeSystem().typeForName(
						MagicClass.DEFAULT_MAGIC_CLASS), name);
				ParsedClassType classType = (ParsedClassType) container;
				classType.fields().add(temp);
				return temp;
			} else {
//				if (name.equals("refqueue") && container.toString().equals("java.lang.Object")) {
//					System.out.println();
//				}
				if (finalRequest) {
					// XXX HAck: this was a wrong class, but...
					FieldInstance temp = new FieldInstance_c(container.typeSystem(), Position
							.compilerGenerated(), container, Flags.PUBLIC.set(Flags.FINAL).set(
							Flags.STATIC), container.typeSystem().typeForName(
							MagicClass.DEFAULT_MAGIC_CLASS), name);
					ParsedClassType classType = (ParsedClassType) container;
					classType.fields().add(temp);
					return temp;
				}
				throw new NoMemberException(NoMemberException.FIELD, "Field \"" + name
						+ "\" not found in type \"" + container + "\".");
			}
		}
		// There are acceptable fields
		else {
			Iterator i = fields.iterator();
			fi = (FieldInstance) i.next();

			// Should not happen
			if (i.hasNext() && allFromMagicContainer(fields)) {
				FieldInstance fi2 = (FieldInstance) i.next();
				throw new SemanticException("Field \"" + name
						+ "\" is ambiguous; it is defined in both " + fi.container() + " and "
						+ fi2.container() + ".");
			}
		}

		// Last check: should never be thrown or we have a real compilation error!
		if (currClass != null && !isAccessible(fi, currClass)) {
			throw new SemanticException("Cannot access " + fi + ".");
		}

		return fi;
	}

	public Set findMemberClasses(ClassType container, String name, boolean finalRequest)
			throws SemanticException {
		Set set = new HashSet();

		try {
			set = super.findMemberClasses(container, name, finalRequest);
		} catch (SemanticException se) {
			if (finalRequest && TypeFactUtil.isStrictMagicHierarchy(container)) {
				set = new HashSet();
			} else {
				throw se;
			}
		} catch (InternalCompilerError ie) {
			if (finalRequest && TypeFactUtil.isStrictMagicHierarchy(container)) {
				set = new HashSet();
			} else {
				throw new SemanticException(ie);
			}
		}

		if (finalRequest && set.size() == 0) {
			MagicClass mc = TypeFactUtil.getFirstMagicClassInHierarchy(container);
			if (mc != null) {
				MagicClass newClass = new MagicClass(name, container.package_().fullName(),
						container.fullName() + "." + name, container.typeSystem());
				newClass.setContainer(container);
				mc.memberClasses().add(newClass);
				set.add(newClass);
			}
		}

		return set;
	}

	@Override
	public MethodInstance findMethod(ReferenceType container, java.lang.String name, List argTypes,
			ClassType currClass) throws SemanticException {
		return findMethod(container, name, argTypes, currClass, null);
	}

	public MethodInstance findMethod(ReferenceType container, java.lang.String name, List argTypes,
			ClassType currClass, Type expectedReturnType) throws SemanticException {
		List tempAcceptable = findAcceptableMethods(container, name, argTypes, currClass);
		List<MethodInstance> acceptable = new ArrayList<MethodInstance>();

		
		// In collective mode, there can be a mismatch in the parameters, try to find the most plausible method
		if (tempAcceptable.size() == 0 && !TypeFactUtil.hasMagicInHierarchy(container)) {
			tempAcceptable = forceFindAcceptableMethods(container, name, argTypes, currClass);
		}
		
		boolean withExpected = expectedReturnType != null;
		if (!withExpected) {
			expectedReturnType = container.typeSystem().typeForName(MagicClass.DEFAULT_MAGIC_CLASS);
			acceptable = tempAcceptable;
		} else {
			acceptable = getAcceptable(tempAcceptable, expectedReturnType);
		}

		MethodInstance temp = null;

		// This is the case when:
		// 1- Calling known methods with unknown parameters.
		// 2- Calling unknown methods from unknown classes.
		if (acceptable.size() == 0) {
			Type firstMagicInterface = getMagicInterface(container);
			MagicClass magicType = TypeFactUtil.getFirstMagicClassInHierarchy(container);

			// Method from a class with a magic class in its hierarchy!
			if (magicType != null) {
				temp = new MagicMethodInstance_c(magicType.typeSystem(), Position
						.compilerGenerated(), magicType.toReference(), Flags.PUBLIC,
						expectedReturnType, name, cleanArgsFromAnon(argTypes), new ArrayList());
				magicType.getMethods().add(temp);
			}
			// Method called from a magic interface: this is possible if the class is abstract.
			else if (firstMagicInterface != null) {
				temp = new MagicMethodInstance_c(firstMagicInterface.typeSystem(), Position
						.compilerGenerated(), firstMagicInterface.toReference(), Flags.PUBLIC,
						expectedReturnType, name, cleanArgsFromAnon(argTypes), new ArrayList());
				MagicClass magicClass = (MagicClass) firstMagicInterface;
				magicClass.getMethods().add(temp);
//				System.err.println("Used a magic interface!!");
			}
		}

		MethodInstance mi = null;

		if (temp != null) {
			mi = temp;
		} else {
			// This is actually the common case when we found some acceptable methods.
			// TODO What happens if we get a method with a type different then the expected one?
			if (acceptable.size() != 0) {
				mi = (MethodInstance) findProcedure(acceptable, container, argTypes, currClass);
			} else {
				// Try to get the error queue...
				// System.err.println("Could not find method" + name + " in " + tempAcceptable
				// + " for expected return type: " + expectedReturnType);
			}
		}

		if (mi == null && acceptable.size() > 0) {
			// Ambiguous, so take the first acceptable method if mi or an
			// argTypes is a magic class.

			// Ambiguous, probably because of redundant magic types and magic methods.
			// TODO Get the most probable acceptable one (create a better filter).
			if (hasMagicMethodContainers(acceptable) || containsMagicClass(argTypes)) {
				mi = acceptable.get(0);
				System.err.println("Still not really good...");
			} else {
				// Skip bug for now...
				if (false) {
					throw new SemanticException("Reference to " + name
							+ " is ambiguous, multiple methods match: " + acceptable);
				} else {
					System.err.println("We got an error here!");
					mi = acceptable.get(0);
				}
			}
		}

		// This is actually an inference and should be documented as such!!!
		// TODO This does not work well if we selected a random method previously...
		// Before, it was an else of the previous if, but we still get the problem of bad filter.
		// So first, create a better filter, THAN, do a cleanup.
		if (withExpected && mi != null && TypeFactUtil.isAnonymousMagicClass(mi.returnType())) {
			// Replace the method!
			ReferenceType rType = mi.container();
			if (rType instanceof MagicClass) {
				MethodInstance newMi = new MagicMethodInstance_c(mi.typeSystem(), mi.position(),
						rType, mi.flags(), expectedReturnType, mi.name(), mi.formalTypes(), mi
								.throwTypes());
				rType.methods().remove(mi);
				rType.methods().add(newMi);
				mi = newMi;
			}
		}

		if (mi == null) {
			if (container instanceof ParsedClassType) {
				ParsedClassType miContainer = (ParsedClassType) container;
				ClassType cType = miContainer.outer();
				if (cType != null) {
					if (withExpected) {
						mi = findMethod(cType.toReference(), name, argTypes, currClass,
								expectedReturnType);
					} else {
						mi = findMethod(cType.toReference(), name, argTypes, currClass);
					}
				}
			}
		}

		return mi;
	}

	private List forceFindAcceptableMethods(ReferenceType container, String name, List argTypes,
			ClassType currClass) throws SemanticException {

		assert_(container);
		assert_(argTypes);

		// The list of acceptable methods. These methods are accessible from
		// currClass, the method call is valid, and they are not overridden
		// by an unacceptable method (which can occur with protected methods
		// only).
		List acceptable = new ArrayList();

		// A list of unacceptable methods, where the method call is valid, but
		// the method is not accessible. This list is needed to make sure that
		// the acceptable methods are not overridden by an unacceptable method.
		List unacceptable = new ArrayList();

		Set visitedTypes = new HashSet();

		LinkedList typeQueue = new LinkedList();
		typeQueue.addLast(container);

		while (!typeQueue.isEmpty()) {
			Type type = (Type) typeQueue.removeFirst();

			if (visitedTypes.contains(type)) {
				continue;
			}

			visitedTypes.add(type);

			if (Report.should_report(Report.types, 2))
				Report.report(2, "Searching type " + type + " for method " + name + "("
						+ listToString(argTypes) + ")");

			if (!type.isReference()) {
				throw new SemanticException("Cannot call method in " + " non-reference type "
						+ type + ".");
			}

			for (Iterator i = type.toReference().methods().iterator(); i.hasNext();) {
				MethodInstance mi = (MethodInstance) i.next();

				if (Report.should_report(Report.types, 3))
					Report.report(3, "Trying " + mi);

				if (mi.name().equals(name) && mi.formalTypes().size() == argTypes.size()) {
					if (isAccessible(mi, currClass)) {
						if (Report.should_report(Report.types, 3)) {
							Report.report(3, "->acceptable: " + mi + " in " + mi.container());
						}

						acceptable.add(mi);
					} else {
						// method call is valid, but the method is
						// unacceptable.
						unacceptable.add(mi);
					}
				}
			}
			if (type.toReference().superType() != null) {
				typeQueue.addLast(type.toReference().superType());
			}

			typeQueue.addAll(type.toReference().interfaces());
		}

		// remove any method in acceptable that are overridden by an
		// unacceptable
		// method.
		for (Iterator i = unacceptable.iterator(); i.hasNext();) {
			MethodInstance mi = (MethodInstance) i.next();
			acceptable.removeAll(mi.overrides());
		}

		return refineAcceptables(acceptable,argTypes);
	}
	
	/**
	 * <p>
	 * This method is overridden to allow public access.
	 * </p>
	 */
	@Override
	public ProcedureInstance findProcedure(List acceptable, ReferenceType container, List argTypes,
			ClassType currClass) throws SemanticException {
		return super.findProcedure(acceptable, container, argTypes, currClass);
	}

	/**
	 * <p>
	 * Returns a list of methods that return a given return type.
	 * </p>
	 * 
	 * <p>
	 * If no methods return the expected return type (implicit cast is not valid), check for methods
	 * with related return types or types that might be related.
	 * </p>
	 */
	public List<MethodInstance> getAcceptable(List<MethodInstance> tempAcceptable,
			Type expectedReturnType) {
		List<MethodInstance> acceptable = new ArrayList<MethodInstance>();
		List<MethodInstance> alsoAcceptable = new ArrayList<MethodInstance>();

		if (expectedReturnType != null) {

			for (Object miObject : tempAcceptable) {
				MethodInstance tempMi = (MethodInstance) miObject;
				Type returnType = tempMi.returnType();

				if (tempMi.returnType().equalsImpl(expectedReturnType)
						|| isImplicitCastValid(returnType, expectedReturnType)) {
					acceptable.add(tempMi);
				} else if (TypeFactUtil.relatedTypes(returnType, expectedReturnType)
						|| TypeFactUtil.mightBeRelated(returnType, expectedReturnType)) {
					alsoAcceptable.add(tempMi);
				}
			}

			if (acceptable.size() == 0 && alsoAcceptable.size() > 0) {
				acceptable = alsoAcceptable;
			}
		}

		return acceptable;
	}

	/**
	 * 
	 * @param container
	 * @return Get the first interface from a container's hierarchy that is a magic class.
	 */
	private Type getMagicInterface(ReferenceType container) {

		Type firstMagicInterface = null;
// XXX Fix this method who was probably never returning a magic interface before!!!
		if (container instanceof MemberInstance) {
			if (container instanceof MagicClass) {
				firstMagicInterface = container.toType();
			} else {
				for (Object interfaceObj : container.interfaces()) {
					Type interfaceType = (Type) interfaceObj;
					if (interfaceType instanceof MagicClass) {
						firstMagicInterface = interfaceType;
						break;
					} else {
						firstMagicInterface = getMagicInterface(interfaceType.toReference());
						if (firstMagicInterface != null) {
							break;
						}
					}
				}
			}
		}
		return firstMagicInterface;
	}

	/**
	 * 
	 * @param acceptable
	 * @return True if one of the method in the list has a magic container.
	 */
	private boolean hasMagicMethodContainers(List acceptable) {
		boolean isMagicClass = false;

		for (Object temp : acceptable) {
			MethodInstance mi = (MethodInstance) temp;
			if (mi.container() instanceof MagicClass) {
				isMagicClass = true;
				break;
			}
		}

		return isMagicClass;
	}

	public boolean isInitialized() {
		return super.systemResolver != null;
	}

	public MagicPrimitive magicBoolean() {
		return MAGIC_BOOLEAN;
	}

	public MagicPrimitive magicInt() {
		return MAGIC_INT;
	}

	public MagicClass magic() {
		MagicClass mc = null;
		try {
			mc = (MagicClass) super.forName(MagicClass.DEFAULT_MAGIC_CLASS);
		} catch (SemanticException se) {
			// FIXME Do something with the exception.
			se.printStackTrace();
		}
		return mc;
	}

	@Override
	public boolean moreSpecific(ProcedureInstance p1, ProcedureInstance p2) {
		// Problem: if p1 and p2 contains MagicClass, both are more specific, so it should be false.

		boolean moreSpecific1 = super.moreSpecific(p1, p2);

		if (moreSpecific1) {
			moreSpecific1 = !super.moreSpecific(p2, p1);
		}

		return moreSpecific1;
	}

	/**
	 * <p>
	 * Will return a list that favors related magic and normal types over unrelated normal types
	 * </p>
	 * 
	 * <p>
	 * For example, if the list {MyMagic1,MyMagic2} is required and those two methods are provided,
	 * the first one will be kept, the second one will be removed from the list:
	 * </p>
	 * 
	 * <ol>
	 * <li>{MyMagic1,RelatedNormalType1}</li>
	 * <li>{MyMagic1,UnrelatedNormalType2]</li>
	 * </ol>
	 * 
	 * @param acceptables
	 * @param argTypes
	 * @return
	 */
	private List<MethodInstance> refineAcceptables(List<MethodInstance> acceptables, List argTypes) {
		List<MethodInstance> refined = new ArrayList<MethodInstance>();
		int size = argTypes.size();
		for (MethodInstance mi : acceptables) {
			List args = mi.formalTypes();
			for (int i = 0; i < size; i++) {
				Type requiredArg = (Type) argTypes.get(i);
				Type providedArg = (Type) args.get(i);
				if (equals(requiredArg, providedArg)
						|| TypeFactUtil.relatedTypes(requiredArg, providedArg)) {
					refined.add(mi);
					break;
				}
			}
		}

		// In case we could not refine it, return the original list.
		if (refined.size() == 0) {
			refined = acceptables;
		}

		return refined;
	}

	/**
	 * Only called when the type system is already initialized. Will inform the system resolver that
	 * there is a new extension info without throwing away the cache.
	 */
	public void reInitialize(ExtensionInfo extInfo) {
		CachingResolver cResolver = (CachingResolver) super.systemResolver;
		cResolver.setExtInfo(extInfo);
	}

}
